/**********************************************************************
 @C-file{
    author              = "Nelson H. F. Beebe",
    version             = "3.2",
    date                = "23 May 1998",
    time                = "18:10:27 MDT",
    filename            = "chkdelim.c",
    address             = "Center for Scientific Computing
			   Department of Mathematics
			   University of Utah
			   Salt Lake City, UT 84112
			   USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 581 4148",
    checksum            = "20527 1669 4671 42594",
    email               = "beebe@math.utah.edu, beebe@acm.org,
			   beebe@ieee.org (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "delimiter check",
    supported           = "yes",
    docstring           = "This program checks for balanced delimiters
			   (braces, angle brackets, parentheses, square
			   brackets, TeX dollars) in one or more files,
			   of if none are specified, then in stdin.

			   chkdelim returns a successful status code
			   (0 on UNIX) if no errors were found, and
			   otherwise, returns an unsuccessful status
			   code (1 on UNIX).

			   Usage:
			       chkdelim [-?] [-author] [-BibTeX] [-copyright]
					[-error-file filename] [-help]
					[-ignore delimiters] [-Lisp] [-nested]
					[-no-BibTeX] [-no-Lisp] [-no-nested]
					[-no-Nroff] [-no-page-reset]
					[-no-paragraph-reset] [-no-Scribe]
					[-no-TeX] [-no-TeXinfo] [-Nroff]
					[-output-file filename] [-page-reset]
					[-paragraph-reset] [-Scribe] [-TeX]
					[-TeXinfo] [-version]
					[files] >outfile

			   Consult the accompanying manual pages
			   for documentation of these options.

			   The source code has been written for a high
			   degree of portability, and should be
			   compilable WITHOUT ANY CHANGES WHATSOEVER on
			   K&R C, ANSI/ISO Standard C, and C++ systems.
			   Only four standard header files are required,
			   and only these standard C library functions
			   are used: exit(), fclose(), fprintf(),
			   fputs(), freopen(), memcpy(), perror(),
			   printf(), sprintf(), strchr(), strlen(),
			   strncpy(), and tolower().  There is no use of
			   dynamic memory allocation, and thus, there
			   are no memory leaks.

			   These symbols may be defined at compile time
			   if needed to adjust to compiler facilities:

			   HAVE_BOOL         C++ compilation with support for
					     bool data type
			   HAVE_SHORT_NAMES  names unique in first 6 characters

			   These symbols may be defined at compile time
			   if needed to adjust array sizes:

			   MAX_PUSHBACK      size of input pushback buffer
			   MAX_STACK         size of delimiter stacks
			   MAX_TEXT          size of input lookahead buffer

			   The checksum field above contains a CRC-16
			   checksum as the first value, followed by the
			   equivalent of the standard UNIX wc (word
			   count) utility output of lines, words, and
			   characters.  This is produced by Robert
			   Solovay's checksum utility.",
 }
 **********************************************************************/
/* -*-C-*- chkdelim - check delimiter balance */
/* /u/sy/beebe/src/lex/chkdelim/chkdelim.c, Wed Apr 24 17:44:22 1991 */
/* Edit by Nelson H.F. Beebe <beebe@plot79.math.utah.edu> */

#define PROGRAM "chkdelim"
#define VERSION "3.2 [23-May-1998]"	/* NB: remember to change this! */

/**********************************************************************/
/* These MAX_xxx default values can be overridden at compile time: */

#if !defined(MAX_PUSHBACK)
#define MAX_PUSHBACK		2048		/* pushback buffer size */
#endif

#if !defined(MAX_STACK)
#define MAX_STACK		1023
#endif

#if !defined(MAX_TEXT)
#define MAX_TEXT		MAX_PUSHBACK /* need not be bigger than this */
#endif

/**********************************************************************/

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#include "args.h"

#if defined(__GNUG__) && defined(sun)
/* g++ on SunOS 4.x is missing these declarations */
extern "C" {
extern char *	strchr ARGS((const char *s, int c));
extern size_t	strlen ARGS((const char *s));
extern char *	strncpy ARGS((char *s1, const char *s2, size_t n));
}
#endif

/***********************************************************************
 Naming conventions:
	* File global variables and enum constants are capitalized.
	* Function names are lowercase.
	* Local variables are lowercase.
	* Preprocessor and compile-time constants are uppercase.
	* Private type names end in _t.
 Programming conventions:
	* Functions are in alphabetical order (except inline ones, which
	  must appear before their first use).
	* Declarations and assignments are in alphabetical order
	  whenever possible.
	* Two blank lines appear between functions.
	* Braces rate their own lines.
	* Explicit typecasting is used where needed.
***********************************************************************/

#if defined(HAVE_SHORT_NAMES)
/* Generate names that are unique in the first 6 characters, if needed */
#define DollarDollar			DDollar
#define Ignore_Angle			Ign_Angle
#define Ignore_At			Ign_At
#define Ignore_Backslash		Ign_Backslash
#define Ignore_Brace			Ign_Brace
#define Ignore_Bracket			Ign_Bkt
#define Ignore_Close_Quote		Ign_Close_Quote
#define Ignore_Dollar			Ign_Dollar
#define Ignore_Embedded_Apostrophe	Ign_Embedded_Apostrophe
#define Ignore_Open_Quote		Ign_Open_Quote
#define Ignore_Paren			Ign_Paren
#define Ignore_Quotation_Mark		Ign_Quotation_Mark
#define Input_Pushback_Buffer		Inp_PB
#define Input_Pushback_Count		Inp_N
#define Mode_TeXinfo			Modetxi
#define Position_Style_t		Pos_S_t
#define Position_t			Pos_t
#define char_close_angle		ccl_angle
#define char_close_brace		ccl_brace
#define char_close_bracket		ccl_square
#define char_close_parenthesis		ccl_parenthesis
#define char_close_quote		ccl_quote
#define char_open_angle			cop_angle
#define char_open_brace			cop_brace
#define char_open_bracket		cop_square
#define char_open_parenthesis		cop_parenthesis
#define char_open_quote			cop_quote
#define check_balance			che_balance
#define check_file			che_file
#define delimiters			delims
#define initialize_action_table		iniact
#define initialize_flags		iniflg
#define initialize_position		inipos
#define initialize_stacks		inistk
#define inline_math			inl_math
#define opt_TeXinfo			opt_txi
#define opt_no_BibTeX			ono_BibTeX
#define opt_no_Lisp			ono_Lisp
#define opt_no_Nroff			ono_Nroff
#define opt_no_Scribe			ono_Scribe
#define opt_no_TeX			ono_TeX
#define opt_no_TeXinfo			ono_txi
#define opt_no_page			ono_pge
#define opt_no_paragraph		ono_paragraph
#define opt_page			opt_pge
#endif

#if !( defined(__GNUG__) || defined(__cplusplus) || defined(c_plusplus) || defined(inline) )
#define inline
#endif

#define ARG_MATCH(s,t,n)	(arg_match(s,t,MAX((n),(int)strlen(s))))
#define	ERROR_PREFIX	"??"	/* this prefixes all error messages */

#ifdef MAX
#undef MAX
#endif
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))

static const int CHAR_ASTERISK		= ((int)'*');
static const int CHAR_AT		= ((int)'@');
static const int CHAR_BACKSLASH		= ((int)'\\');
static const int CHAR_BLANK		= ((int)' ');
static const int CHAR_CLOSE_ANGLE	= ((int)'>');
static const int CHAR_CLOSE_BRACE	= ((int)'}');
static const int CHAR_CLOSE_BRACKET	= ((int)']');
static const int CHAR_CLOSE_PARENTHESIS	= ((int)')');
static const int CHAR_CLOSE_QUOTE	= ((int)'\'');
static const int CHAR_CR		= ((int)'\r');
static const int CHAR_DOLLAR		= ((int)'$');
static const int CHAR_EXCLAMATION	= ((int)'!');
static const int CHAR_FF		= ((int)'\f');
static const int CHAR_HT		= ((int)'\t');
static const int CHAR_MINUS		= ((int)'-');
static const int CHAR_NEWLINE		= ((int)'\n');
static const int CHAR_NUL		= ((int)'\0');
static const int CHAR_OPEN_ANGLE	= ((int)'<');
static const int CHAR_OPEN_BRACE	= ((int)'{');
static const int CHAR_OPEN_BRACKET	= ((int)'[');
static const int CHAR_OPEN_PARENTHESIS	= ((int)'(');
static const int CHAR_OPEN_QUOTE	= ((int)'`');
static const int CHAR_QUERY		= ((int)'?');
static const int CHAR_QUOTATION_MARK	= ((int)'"');
static const int CHAR_SLASH		= ((int)'/');

/***********************************************************************
POSITION_FORMAT_VERBOSE is not used here, but its pieces appear in calls
to sprintf() in get_position().  POSITION_SIZE reflects the maximum size
needed for such a formatted string.

The magic number 1024 in the definition below is the limit on a filename
length.  An implementation-defined symbolic name is not used here,
because there is wide variation in the name of this limit in old K&R C
systems before ANSI/ISO Standard C named it FILENAME_MAX; I have found
at least these: FNMAX, MAXPATH, MAXPATHLEN, PATH_MAX, and _MAX_PATH,
plus the Standard FILENAME_MAX.

In a POSIX environment, or in a network-mounted file system environment,
this value may differ between file systems on the same host, and I do
not want to have to evaluate

    pathconf(pathname,_PC_NAME_MAX) + 1

because that would require including <unistd.h>, and older UNIX C
implementations, and many non-UNIX C and C++ ones, lack POSIX support.

Neither do I want to do dynamic memory allocation based on the lengths
of the filename arguments passed to this function, because that would
involve calling malloc(), which is not otherwise needed here.

All that we really need here is the MAXIMUM value possible in current
file systems, and that is the magic number 1024, which in practice is
likely to be much larger than ever required by users of this program,
who for reasons of output readability would be unlikely to specify
absolute paths to argument files.

For safety, when File_Name is assigned in main(), it is truncated to
MAX_FILEPATH if necessary.  That way, it is impossible for a malicious
user to cause an internal buffer overrun error by passing a very long
filename argument; such errors have been the commonest source of
security loopholes on the Internet. */

#define MAX_FILEPATH	1024

#define NOOP

#define POSITION_FORMAT_VERBOSE "%.1024s:%ld:%ld:%ld:.-%ld"
#define POSITION_SIZE	(MAX_FILEPATH + 1 + 20 + 1 + 20 + 1 + 20 + 1 + 2 + 20 + 1)
		/* NB: 20 == maximum length of 64-bit signed integer */

#ifdef FALSE
#undef FALSE
#endif

#ifdef TRUE
#undef TRUE
#endif

#if defined(__cplusplus) && defined(HAVE_BOOL)
typedef bool		Boolean_t;
static const int FALSE = false;
static const int TRUE  = true;
#else
typedef int		Boolean_t;
#define FALSE	0
#define TRUE	1
#endif

#if !defined(UCHAR_MAX)
#define UCHAR_MAX 255
#endif

typedef enum
{
    Mode_ASCII = 0,
    Mode_BibTeX = 1,
    Mode_Lisp = 2,
    Mode_Nroff = 3,
    Mode_Scribe = 4,
    Mode_TeX = 5,
    Mode_TeXinfo = 6
} Mode_t;

typedef struct
{
    long char_number;
    long column_number;
    long line_number;
    int delimiter;
} Position_t;

typedef enum
{
    Compact = 0,
    Verbose = 1
} Position_Style_t;

typedef struct {
    Position_t stack[MAX_STACK];
    size_t top;			/* index of first unused slot in stack[] */
} Stack_t;

typedef unsigned int	Uint_t;

/* These objects are static for privacy, and being file global, they
   are guaranteed to be pre-zeroed, which for stacks means they are
   empty. */

static Boolean_t	Ignore_Angle = FALSE;
static Boolean_t	Ignore_At = TRUE;
static Boolean_t	Ignore_Backslash = FALSE;
static Boolean_t	Ignore_Brace = FALSE;
static Boolean_t	Ignore_Bracket = FALSE;
static Boolean_t	Ignore_Close_Quote = FALSE;
static Boolean_t	Ignore_Dollar = TRUE;
static Boolean_t	Ignore_Embedded_Apostrophe = FALSE;
static Boolean_t	Ignore_Formfeed = TRUE;
static Boolean_t	Ignore_Open_Quote = FALSE;
static Boolean_t	Ignore_Paren = FALSE;
static Boolean_t	Ignore_Quotation_Mark = FALSE;
static Boolean_t	In_Math = FALSE;
static Boolean_t	In_Quoted_String = FALSE;
static Boolean_t	Nested = TRUE;
static Boolean_t	Paragraph_Reset = FALSE;
static Boolean_t	TeX_quotes = FALSE;

static const char *	File_Name = (const char *) NULL;
static const char *	Program_Name = PROGRAM;

static FILE *		Stderr;
static FILE *		Stdout;

static int		Error_Count = 0;
static int		Last_c = 0;
static int 		Input_Pushback_Buffer[MAX_PUSHBACK];
static int 		Input_Pushback_Count = 0;

static Mode_t		Mode = Mode_ASCII;

static Position_t	Current_Position;
static Position_t	Last_Newline_Position;

static Stack_t		Angle;
static Stack_t		Brace;
static Stack_t		Bracket;
static Stack_t		Close_Quote;
static Stack_t		Dollar;
static Stack_t		DollarDollar;
static Stack_t		Dstack;
static Stack_t		Open_Quote;
static Stack_t		Parenthesis;
static Stack_t		Quotation_Mark;

static void		(*Action[UCHAR_MAX+1]) ARGS((void));

int			main ARGS((int argc, char *argv[]));

static int		arg_match ARGS((const char *option,
					const char *argname, int matchlen));
static void		char_at ARGS((void));
static void		char_close_angle ARGS((void));
static void		char_close_brace ARGS((void));
static void		char_close_bracket ARGS((void));
static void		char_close_parenthesis ARGS((void));
static void		char_close_quote ARGS((void));
static void		char_dollar ARGS((void));
static void		char_formfeed ARGS((void));
static void		char_open_angle ARGS((void));
static void		char_open_brace ARGS((void));
static void		char_open_bracket ARGS((void));
static void		char_open_parenthesis ARGS((void));
static void		char_open_quote ARGS((void));
static void		char_quotation_mark ARGS((void));
static void		check_balance ARGS((const char *end_name));
static void		check_file ARGS((void));
static void		display_math ARGS((void));
static void		do_file ARGS((void));
static void		dump ARGS((Stack_t *stack, const char *delimiter));
static const char *	get_position ARGS((Stack_t *stack, size_t stack_index,
					   Position_Style_t style));
static int		have_backslash ARGS((int c));
static int		input ARGS((void));
static void		have_dollar ARGS((void));
static void		have_newline ARGS((void));
static void		initialize_action_table ARGS((void));
static void		initialize_flags ARGS((void));
static void		initialize_position ARGS((void));
static void		initialize_stacks ARGS((void));
static void		initialize_startup_defaults ARGS((void));
static void		inline_math ARGS((void));

#if defined(__VMS__)
#define isoptionchar(c) (((int)(c) == CHAR_SLASH) || ((int)(c) == CHAR_MINUS))
#else
#if defined(__MSDOS__) || defined(_MSDOS) || defined(MSDOS) || \
	defined(DOS) || defined(__TURBOC__) || defined(TURBOC)
#define isoptionchar(c) (((int)(c) == CHAR_SLASH) || ((int)(c) == CHAR_MINUS))
#else
#define isoptionchar(c) ((c) == CHAR_MINUS)
#endif
#endif

static void		new_paragraph ARGS((void));
static void		opt_BibTeX ARGS((void));
static void		opt_Lisp ARGS((void));
static void		opt_Nroff ARGS((void));
static void		opt_Scribe ARGS((void));
static void		opt_TeX ARGS((void));
static void		opt_TeXinfo ARGS((void));
static void		opt_author ARGS((void));
static void		opt_copyright ARGS((void));
static void		opt_error_file ARGS((const char *filename));
static void		opt_help ARGS((void));
static void		opt_ignore ARGS((const char *delimiters));
static void		opt_nested ARGS((void));
static void		opt_no_BibTeX ARGS((void));
static void		opt_no_Lisp ARGS((void));
static void		opt_no_Nroff ARGS((void));
static void		opt_no_Scribe ARGS((void));
static void		opt_no_nested ARGS((void));
static void		opt_no_TeX ARGS((void));
static void		opt_no_TeXinfo ARGS((void));
static void		opt_no_page ARGS((void));
static void		opt_no_paragraph ARGS((void));
static void		opt_output_file ARGS((const char *filename));
static void		opt_page ARGS((void));
static void		opt_paragraph ARGS((void));
static void		opt_version ARGS((void));
static void		ordinary_char ARGS((void));
static void		out_strings ARGS((FILE *fpout, const char *lines[]));
static int		peek ARGS((void));
static void		pop ARGS((Stack_t *stack, int expect));
static void		push ARGS((Stack_t *stack, int c));
static void		reset ARGS((size_t keep));
static int		set_position ARGS((int c));
static void		show_version ARGS((void));

/* Some C implementations already have strnicmp(), possibly with a
different prototype than ours, so we rename ours with this
definition. */
#define strnicmp			Strnic

static int		strnicmp ARGS((const char *s1, const char *s2,
				       size_t n));

static int		unput ARGS((int));
static void		usage ARGS((const char *bad_option));


inline
static int
input(VOID)
{
    return (set_position((Input_Pushback_Count > 0) ?
			 Input_Pushback_Buffer[--Input_Pushback_Count] :
			 getchar()));
}


inline
static void
ordinary_char(VOID)
{
}


inline
static int
peek(VOID)
{
    return (unput(input()));
}


static int
#if NEW_STYLE
arg_match(const char *option, const char *argname, int matchlen)
#else
arg_match(option,argname,matchlen)
const char	*option;		/* command-line option */
const char	*argname;		/* "-optionname" */
int		matchlen;		/* number of chars to match */
#endif
{
    if (isoptionchar(option[0]) && (option[0] == option[1]))
	++option;			/* handle GNU/POSIX --option */
    return ((matchlen == 0) ? 0 :
	    (strnicmp(option+1,argname+1,(size_t)(matchlen-1)) == 0));
}


static void
char_at(VOID)
{
    if ((Mode == Mode_BibTeX) && (Current_Position.column_number == 1))
	reset(0);		/* reset delimiter stack at new BibTeX entry */
}


static void
char_close_angle(VOID)
{
    pop(&Angle,CHAR_OPEN_ANGLE);
    pop(&Dstack,CHAR_OPEN_ANGLE);
}


static void
char_close_brace(VOID)
{
    pop(&Brace,CHAR_OPEN_BRACE);
    pop(&Dstack,CHAR_OPEN_BRACE);
}


static void
char_close_bracket(VOID)
{
    pop(&Bracket,CHAR_OPEN_BRACKET);
    pop(&Dstack,CHAR_OPEN_BRACKET);
}


static void
char_close_parenthesis(VOID)
{
    pop(&Parenthesis,CHAR_OPEN_PARENTHESIS);
    pop(&Dstack,CHAR_OPEN_PARENTHESIS);
}


static void
char_close_quote(VOID)
{
    if (Ignore_Embedded_Apostrophe)
    {
	if (isalpha(Last_c) && isalpha(peek()))
	    return;
	else if (isspace(Last_c) && isdigit(peek()))
	    return;
    }

    if ((Mode == Mode_Lisp) &&
	(isspace(Last_c) || (Last_c == CHAR_OPEN_PARENTHESIS)))
	return;				/* ignore (foo 'bar) quoting  */

    /* Open and close quotes inside math mode are ignored. */

    if (!In_Math)
    {
	if (TeX_quotes)
	{
	    pop(&Dstack,    CHAR_OPEN_QUOTE);
	    pop(&Open_Quote,CHAR_OPEN_QUOTE);
	}
	else if ((Close_Quote.top % 2) == 0)	/* expect open delimiter */
	{
	    push(&Dstack,     CHAR_CLOSE_QUOTE);
	    push(&Close_Quote,CHAR_CLOSE_QUOTE);
	}
	else				/* expect close delimiter */
	{
	    pop(&Dstack,     CHAR_CLOSE_QUOTE);
	    pop(&Close_Quote,CHAR_CLOSE_QUOTE);
	}
    }
}


static void
char_dollar(VOID)
{
    /* TO DO: fix this: it fails for $$ .. $$ style input! */
    if (In_Math)
	pop(&Dstack,CHAR_DOLLAR);
    else
	push(&Dstack,CHAR_DOLLAR);
    In_Math = In_Math ? FALSE : TRUE;
}


static void
char_formfeed(VOID)
{
    if ((Mode == Mode_TeX) && (Dollar.top % 2))
	(void)fprintf(Stdout,"%s:page break in inline math mode\n",
		     get_position((Stack_t*)NULL,0,Verbose));
    if ((Mode == Mode_TeX) && (DollarDollar.top % 2))
	(void)fprintf(Stdout,"%s:page break in display math mode\n",
		     get_position((Stack_t*)NULL,0,Verbose));
    check_balance("page");
}


static void
char_open_angle(VOID)
{
    push(&Angle,CHAR_OPEN_ANGLE);
    push(&Dstack,CHAR_OPEN_ANGLE);
}


static void
char_open_brace(VOID)
{
    push(&Brace,CHAR_OPEN_BRACE);
    push(&Dstack,CHAR_OPEN_BRACE);
}


static void
char_open_bracket(VOID)
{
    push(&Bracket,CHAR_OPEN_BRACKET);
    push(&Dstack,CHAR_OPEN_BRACKET);
}


static void
char_open_parenthesis(VOID)
{
    push(&Parenthesis,CHAR_OPEN_PARENTHESIS);
    push(&Dstack,CHAR_OPEN_PARENTHESIS);
}


static void
char_open_quote(VOID)
{
    /* Open and close quotes inside math mode are ignored. */

    if (!In_Math)
    {
	if ((Mode == Mode_TeX) &&
	    ((Last_c == CHAR_QUERY) || (Last_c == CHAR_EXCLAMATION)))
	    return;		/* ignore ?` and !` for Spanish ? and ! */
	push(&Dstack,CHAR_OPEN_QUOTE);
	if (TeX_quotes)
	    push(&Open_Quote,CHAR_OPEN_QUOTE);
	else if ((Open_Quote.top % 2) == 0)	/* expect open delimiter */
	    push(&Open_Quote,CHAR_OPEN_QUOTE);
	else				/* expect close delimiter */
	    pop(&Open_Quote,CHAR_OPEN_QUOTE);
    }
}


static void
char_quotation_mark(VOID)
{
    if ((Mode == Mode_BibTeX) && (Brace.top < 2) )
    {		/* quotes at brace level 2 or higher are ordinary characters */
	if (In_Quoted_String)
	    pop(&Dstack,CHAR_QUOTATION_MARK);
	else
	    push(&Dstack,CHAR_QUOTATION_MARK);
	In_Quoted_String = In_Quoted_String ? FALSE : TRUE;
	if (!In_Quoted_String)
	    reset(1);	/* quoted strings lie inside one level of bracing */
    }
    if ((Quotation_Mark.top % 2) == 0)		/* expect open delimiter */
	push(&Quotation_Mark,CHAR_QUOTATION_MARK);
    else				/* expect close delimiter */
	pop(&Quotation_Mark,CHAR_QUOTATION_MARK);
}


static void
#if NEW_STYLE
check_balance(const char *end_name)
#else
check_balance(end_name)
const char *end_name;
#endif
{
    if (Dstack.top > 0)
    {
	(void)fprintf(Stdout,"%s:%d unclosed delimiter%s at end of %s\n",
		     get_position((Stack_t*)NULL,0,Verbose),
		     (int)Dstack.top,
		     ((Dstack.top > 1) ? "s" : ""),
		     end_name);
	/* We don't do "dump(&Dstack,(const char*)NULL);" because
	   that duplicates information from the other stacks */
    }

    if (TeX_quotes)
    {
	if (!Ignore_Open_Quote) dump(&Open_Quote,"`");
    }
    else
    {
	if (!Ignore_Close_Quote) dump(&Close_Quote,"'");
	if (!Ignore_Open_Quote) dump(&Open_Quote,"`");
    }
    if (!Ignore_Dollar)		dump(&Dollar,"$");
    if (!Ignore_Dollar)		dump(&DollarDollar,"$$");
    if (!Ignore_Quotation_Mark) dump(&Quotation_Mark,"\"");
    if (!Ignore_Angle)		dump(&Angle,"<");
    if (!Ignore_Brace)		dump(&Brace,"{");
    if (!Ignore_Paren)		dump(&Parenthesis,"(");
    if (!Ignore_Bracket)	dump(&Bracket,"[");

    initialize_flags();
    initialize_stacks();
}


static void
check_file(VOID)
{
    int c;

    initialize_action_table();
    initialize_flags();
    initialize_stacks();
    initialize_position();

    while ((c = input(), c != EOF))
    {
	if ((Last_c == CHAR_BACKSLASH) && (!Ignore_Backslash))
	    c = have_backslash(c);
	else if ((Last_c == CHAR_AT) &&
		 ((Mode == Mode_Scribe) || (Mode == Mode_TeXinfo)))
	  NOOP; /*EMPTY*/		/* ignore x in @x, for all x */
	else if ((c == CHAR_DOLLAR) && !Ignore_Dollar)
	    have_dollar();
	else if (c == CHAR_NEWLINE)
	    have_newline();
	else
	    (*Action[(Uint_t)c])();
	/* TO DO: can this Action() call ever call check_dollar()? */
	Last_c = c;
    }
}


static void
display_math(VOID)
{
    if ((DollarDollar.top % 2) == 0)	/* expect open delimiter */
	push(&DollarDollar,CHAR_DOLLAR);
    else				/* expect close delimiter */
	pop(&DollarDollar,CHAR_DOLLAR);
}


static void
do_file(VOID)
{
    check_file();
    check_balance("file");
}


static void
#if NEW_STYLE
dump(
Stack_t *stack,
const char *s
)
#else /* NOT NEW_STYLE */
dump(stack,s)			/* show unmatched delimiter positions */
Stack_t *stack;
const char *s;
#endif /* NEW_STYLE */
{
    size_t k;
    char temp[2];

    temp[1] = (char)CHAR_NUL;
    for (k = 0; k < stack->top; ++k)
    {
	temp[0] = (char)stack->stack[k].delimiter;
	(void)fprintf(Stdout,"%s:unmatched delimiter %s\n",
		     get_position(stack,k,Verbose),
		     ((s == (const char*)NULL) ? (const char*)&temp[0] : s));
    }
}


static const char *
#if NEW_STYLE
get_position(Stack_t *stack, size_t stack_index, Position_Style_t style)
#else /* NOT NEW_STYLE */
get_position(stack, stack_index, style)
Stack_t *stack;
size_t stack_index;
Position_Style_t style;
#endif /* NEW_STYLE */
{
    long	offset;
    static char	posbuf[POSITION_SIZE];
    Position_t	position;

    position = (stack == (Stack_t*)NULL) ? Current_Position :
		stack->stack[stack_index];

    offset = Current_Position.char_number - position.char_number;

    posbuf[0] = (char)CHAR_NUL;
    if (style != Compact)
	(void)sprintf(strchr(posbuf,CHAR_NUL), "%.*s:",
		      MAX_FILEPATH, File_Name);
    (void)sprintf(strchr(posbuf,CHAR_NUL), "%ld:", (long)position.line_number);
    (void)sprintf(strchr(posbuf,CHAR_NUL), "%ld:", (long)position.column_number);
    (void)sprintf(strchr(posbuf,CHAR_NUL), "%ld:", (long)position.char_number);
    if (offset != 0)
	(void)sprintf(strchr(posbuf,CHAR_NUL), ".-%ld:", (long)offset);
    *(strchr(posbuf,CHAR_NUL)-1) = (char)CHAR_NUL; /* replace final colon by NUL */

    return ((const char *) &posbuf[0]);
}


static int
#if NEW_STYLE
have_backslash(int c)
#else
have_backslash(c)
int c;
#endif
{
    if ((Mode == Mode_Nroff) && (c == CHAR_ASTERISK)) /* check for "\*(" */
    {
	if (peek() == CHAR_OPEN_PARENTHESIS)
	{			/* troff macro */
	    ordinary_char();
	    c = input();
	}
    }
    ordinary_char();
    return (c);
}


static void
have_dollar(VOID)
{
    if (peek() == CHAR_DOLLAR)
    {
	(void)input();
	display_math();
    }
    else
	inline_math();
}


static void
have_newline(VOID)
{
    int			c;
    size_t		len;
    char		text[MAX_TEXT];

    /* TO DO: Provide definition of is_hspace() that works on Apple
       Macintosh MacOS, where CR, instead of LF, terminates lines. */

#define is_hspace(c) \
		    (((c) == CHAR_BLANK) || \
		     ((c) == CHAR_CR) || \
		     ((c) == CHAR_HT))

    len = 0;
    while ((c = input(), (c != EOF) && is_hspace(c) && (len < MAX_TEXT)))
	text[len++] = (char)c;
    if (c == CHAR_NEWLINE)
	new_paragraph();
    else if (c != EOF)	/* did not match empty line: "\n[ \t\r]*\n" */
    {			/* so restore lookahead */
	while (len > 0)
	    (void)unput((int)text[--len]);
	(*Action[(Uint_t)c])();
    }

#undef is_hspace
}


static void
initialize_action_table(VOID)
{
    size_t k;

    for (k = 0; k <= (size_t)UCHAR_MAX; ++k)
	Action[k] = ordinary_char;

    Action[(Uint_t)CHAR_AT]               = Ignore_At             ? ordinary_char : char_at;
    Action[(Uint_t)CHAR_CLOSE_ANGLE]      = Ignore_Angle          ? ordinary_char : char_close_angle;
    Action[(Uint_t)CHAR_CLOSE_BRACE]      = Ignore_Brace          ? ordinary_char : char_close_brace;
    Action[(Uint_t)CHAR_CLOSE_BRACKET]    = Ignore_Bracket        ? ordinary_char : char_close_bracket;
    Action[(Uint_t)CHAR_CLOSE_PARENTHESIS]= Ignore_Paren          ? ordinary_char : char_close_parenthesis;
    Action[(Uint_t)CHAR_CLOSE_QUOTE]      = Ignore_Close_Quote    ? ordinary_char : char_close_quote;
    Action[(Uint_t)CHAR_DOLLAR]           = Ignore_Dollar         ? ordinary_char : char_dollar;
    Action[(Uint_t)CHAR_FF]               = Ignore_Formfeed       ? ordinary_char : char_formfeed;
    Action[(Uint_t)CHAR_OPEN_ANGLE]       = Ignore_Angle          ? ordinary_char : char_open_angle;
    Action[(Uint_t)CHAR_OPEN_BRACE]       = Ignore_Brace          ? ordinary_char : char_open_brace;
    Action[(Uint_t)CHAR_OPEN_BRACKET]     = Ignore_Bracket        ? ordinary_char : char_open_bracket;
    Action[(Uint_t)CHAR_OPEN_PARENTHESIS] = Ignore_Paren          ? ordinary_char : char_open_parenthesis;
    Action[(Uint_t)CHAR_OPEN_QUOTE]       = Ignore_Open_Quote     ? ordinary_char : char_open_quote;
    Action[(Uint_t)CHAR_QUOTATION_MARK]   = Ignore_Quotation_Mark ? ordinary_char : char_quotation_mark;
}


static void
initialize_flags(VOID)
{
    In_Math = FALSE;
    In_Quoted_String = FALSE;
}


static void
initialize_position(VOID)
{
    Last_c = CHAR_NEWLINE;	/* so first file char == first char of line */
    Current_Position.char_number = 0;
    Current_Position.column_number = 0;
    Current_Position.delimiter = 0;
    Current_Position.line_number = 1;
}


static void
initialize_stacks(VOID)
{
    Angle.top = 0;
    Brace.top = 0;
    Bracket.top = 0;
    Close_Quote.top = 0;
    Dollar.top = 0;
    DollarDollar.top = 0;
    Dstack.top = 0;
    Open_Quote.top = 0;
    Parenthesis.top = 0;
    Quotation_Mark.top = 0;
}


static void
initialize_startup_defaults(VOID)
{
    Ignore_Angle                = FALSE;
    Ignore_At                   = TRUE;
    Ignore_Backslash            = FALSE;
    Ignore_Brace                = FALSE;
    Ignore_Bracket              = FALSE;
    Ignore_Close_Quote          = FALSE;
    Ignore_Dollar               = TRUE;
    Ignore_Embedded_Apostrophe  = FALSE;
    Ignore_Formfeed		= TRUE;
    Ignore_Open_Quote           = FALSE;
    Ignore_Paren                = FALSE;
    Ignore_Quotation_Mark       = FALSE;

    Mode                        = Mode_ASCII;
}


static void
inline_math(VOID)
{
    if ((Dollar.top % 2) == 0)		/* expect open delimiter */
	push(&Dollar,CHAR_DOLLAR);
    else				/* expect close delimiter */
	pop(&Dollar,CHAR_DOLLAR);
}


int
#if NEW_STYLE
main(int argc, char *argv[])
#else /* K&R style */
main(argc, argv)
int argc;
char *argv[];
#endif /* NEW_STYLE */
{
    size_t file_count;
    size_t k;
    FILE *fp;

    Stderr = stderr;			/* need modifiable file handles */
    Stdout = stdout;			/* for stderr and stdout */

    Program_Name = argv[0];		/* NB: used only in fputs() calls, so
					   it CANNOT cause an internal buffer
					   overrun error */

    initialize_startup_defaults();

    file_count = 0;
    for (k = 1; k < (size_t)argc; ++k)
    {
	if (ARG_MATCH(argv[k],"-author",2))
	    opt_author();		/* does not return */
	else if (ARG_MATCH(argv[k],"-BibTeX",2))
	    opt_BibTeX();
	else if (ARG_MATCH(argv[k],"-copyright",2))
	    opt_copyright();		/* does not return */
	else if (ARG_MATCH(argv[k],"-error-file",2))
	    opt_error_file(argv[++k]);
	else if (ARG_MATCH(argv[k],"-help",2) || ARG_MATCH(argv[k],"-?",2))
	    opt_help();			/* does not return */
	else if (ARG_MATCH(argv[k],"-ignore",2))
	    opt_ignore(argv[++k]);
	else if (ARG_MATCH(argv[k],"-Lisp",2))
	    opt_Lisp();
	else if (ARG_MATCH(argv[k],"-nested",3))
	    opt_nested();
	else if (ARG_MATCH(argv[k],"-Nroff",3))
	    opt_Nroff();
	else if (ARG_MATCH(argv[k],"-no-BibTeX",5))
	    opt_no_BibTeX();
	else if (ARG_MATCH(argv[k],"-no-Lisp",5))
	    opt_no_Lisp();
	else if (ARG_MATCH(argv[k],"-no-nested",6))
	    opt_no_nested();
	else if (ARG_MATCH(argv[k],"-no-Nroff",6))
	    opt_no_Nroff();
	else if (ARG_MATCH(argv[k],"-no-page-reset",7))
	    opt_no_page();
	else if (ARG_MATCH(argv[k],"-no-paragraph-reset",7))
	    opt_no_paragraph();
	else if (ARG_MATCH(argv[k],"-no-Scribe",5))
	    opt_no_Scribe();
	else if (ARG_MATCH(argv[k],"-no-TeX",7))
	    opt_no_TeX();
	else if (ARG_MATCH(argv[k],"-no-TeXinfo",8))
	    opt_no_TeXinfo();
	else if (ARG_MATCH(argv[k],"-output-file",2))
	    opt_output_file(argv[++k]);
	else if (ARG_MATCH(argv[k],"-page-reset",4))
	    opt_page();
	else if (ARG_MATCH(argv[k],"-paragraph-reset",4))
	    opt_paragraph();
	else if (ARG_MATCH(argv[k],"-Scribe",2))
	    opt_Scribe();
	else if (ARG_MATCH(argv[k],"-TeX",4))
	    opt_TeX();
	else if (ARG_MATCH(argv[k],"-TeXinfo",5))
	    opt_TeXinfo();
	else if (ARG_MATCH(argv[k],"-version",2))
	    opt_version();		/* does not return */
	else if (isoptionchar(argv[k][0]))
	{
	    usage(argv[k]);
	    exit(EXIT_FAILURE);
	}
	else
	{
	    file_count++;
	    fp = freopen(argv[k],"r",stdin);
	    if (fp == (FILE*)NULL)
	    {
		(void)fprintf(Stderr,
			      "\n%s Ignoring open failure on input file [%s]\n",
			      ERROR_PREFIX, argv[k]);
		(void)perror("perror() says");
	    }
	    else
	    {
		/* Prevent malicious users from causing an internal buffer
		   overrun error: truncate File_Name from the left if
		   necessary. Since the file is already open, and we
		   only use File_Name for output messages, this
		   truncation is harmless. */
		File_Name = (strlen(argv[k]) < MAX_FILEPATH) ? argv[k] :
			    &argv[k][strlen(argv[k]) - MAX_FILEPATH];
		do_file();
		(void)fclose(fp);
	    }
	}
    }
    if (file_count == 0)
    {
	File_Name = "stdin";
	do_file();
    }

    exit((Error_Count > 0) ? EXIT_FAILURE : EXIT_SUCCESS);

    /* NOTREACHED */
    return (0);
}


static void
new_paragraph(VOID)
{
    if ((Mode == Mode_TeX) && (Dollar.top % 2))
	(void)fprintf(Stdout,"%s:empty line in inline math mode\n",
		     get_position((Stack_t*)NULL,0,Verbose));
    if ((Mode == Mode_TeX) && (DollarDollar.top % 2))
	(void)fprintf(Stdout,"%s:empty line in display math mode\n",
		     get_position((Stack_t*)NULL,0,Verbose));
    if (Paragraph_Reset)
	check_balance("paragraph");
}


static void
opt_author(VOID)
{
    static const char *author_text[] =
    {
	"Author:\n",
	"\tNelson H. F. Beebe, Ph.D.\n",
	"\tCenter for Scientific Computing\n",
	"\tUniversity of Utah\n",
	"\tDepartment of Mathematics, 322 INSCC\n",
	"\t155 S 1400 E RM 233\n",
	"\tSalt Lake City, UT 84112-0090\n",
	"\tUSA\n",
	"\tTel: +1 801 581 5254\n",
	"\tFAX: +1 801 585 1640, +1 801 581 4148\n",
	"\tEmail: beebe@math.utah.edu, beebe@acm.org, beebe@ieee.org\n",
	"\tWWW URL: http://www.math.utah.edu/~beebe\n",
	"\n",
	(const char*)NULL,
    };

    out_strings(Stderr, author_text);
    exit(EXIT_SUCCESS);
}


static void
opt_BibTeX(VOID)
{
    opt_TeX();
    Ignore_At = FALSE;
    Ignore_Embedded_Apostrophe = TRUE;
    Mode = Mode_BibTeX;
}


static void
opt_copyright(VOID)
{
    static const char *copyright_text[] =
    {
	"Copyright: none\n",
	"\t*********************************************\n",
	"\t*********************************************\n",
	"\t*** This program is in the PUBLIC DOMAIN. ***\n",
	"\t*********************************************\n",
	"\t*********************************************\n",
	"\n",
	(const char*)NULL,
    };
    out_strings(Stderr, copyright_text);
    exit(EXIT_SUCCESS);
}


static void
#if NEW_STYLE
opt_error_file(
const char *filename
)
#else /* NOT NEW_STYLE */
opt_error_file(filename)
const char *filename;
#endif /* NEW_STYLE */
{
    Stderr = freopen(filename,"w",stderr);
    if (Stderr == (FILE*)NULL)
    {
	Stderr = Stdout;
	(void)fprintf(Stderr,
		      "\n%s open failure on error file [%s]: using stdout instead\n",
		      ERROR_PREFIX, filename);
    }
}


static void
opt_help(VOID)
{
    static const char *help_text[] =
    {
	"On UNIX systems, try\n",
	"\tman ", PROGRAM, "\n",
	"for a detailed description.\n",
	"\n",
	(const char*)NULL,
    };

    usage((const char*)NULL);
    out_strings(Stderr, help_text);
    exit(EXIT_SUCCESS);
}


static void
#if NEW_STYLE
opt_ignore(
const char *delimiters
)
#else /* NOT NEW_STYLE */
opt_ignore(delimiters)
const char *delimiters;
#endif /* NEW_STYLE */
{

#define is_set(x) ((strchr(delimiters,x) == (const char*)NULL) ? FALSE : TRUE)

    if (is_set(CHAR_OPEN_ANGLE) || is_set(CHAR_CLOSE_ANGLE) ||
	is_set('L') || is_set('G') ||
	is_set('l') || is_set('g'))
	Ignore_Angle = TRUE;
    if (is_set(CHAR_BACKSLASH))
	Ignore_Backslash = TRUE;
    if (is_set(CHAR_OPEN_BRACE) || is_set(CHAR_CLOSE_BRACE))
	Ignore_Brace = TRUE;
    if (is_set(CHAR_OPEN_PARENTHESIS) || is_set(CHAR_CLOSE_PARENTHESIS))
	Ignore_Paren = TRUE;
    if (is_set(CHAR_OPEN_BRACKET) || is_set(CHAR_CLOSE_BRACKET))
	Ignore_Bracket = TRUE;
    if (is_set(CHAR_CLOSE_QUOTE))
	Ignore_Close_Quote = TRUE;
    if (is_set(CHAR_DOLLAR))
	Ignore_Dollar = TRUE;
    if (is_set(CHAR_OPEN_QUOTE))
	Ignore_Open_Quote = TRUE;
    if (is_set(CHAR_QUOTATION_MARK))
	Ignore_Quotation_Mark = TRUE;

    if (TeX_quotes)
	Ignore_Close_Quote = Ignore_Open_Quote =
	  (Ignore_Close_Quote || Ignore_Open_Quote);
#undef is_set
}


static void
opt_Lisp(VOID)
{
    initialize_startup_defaults();
    Ignore_Angle = TRUE;
    Ignore_Embedded_Apostrophe = TRUE;
    Ignore_Backslash = FALSE;
    Mode = Mode_Lisp;
}


static void
opt_nested(VOID)
{
    Nested = TRUE;
}


static void
opt_no_BibTeX(VOID)
{
    initialize_startup_defaults();
}


static void
opt_no_Lisp(VOID)
{
    initialize_startup_defaults();
}


static void
opt_no_nested(VOID)
{
    Nested = FALSE;
}


static void
opt_no_Nroff(VOID)
{
    initialize_startup_defaults();
}


static void
opt_no_page(VOID)
{
    Ignore_Formfeed = TRUE;
}


static void
opt_no_paragraph(VOID)
{
    Paragraph_Reset = FALSE;
}


static void
opt_no_Scribe(VOID)
{
    initialize_startup_defaults();
}


static void
opt_no_TeX(VOID)
{
    initialize_startup_defaults();
}


static void
opt_no_TeXinfo(VOID)
{
    initialize_startup_defaults();
}


static void
opt_Nroff(VOID)
{
    initialize_startup_defaults();
    Ignore_At = TRUE;
    Ignore_Backslash = FALSE;
    Ignore_Embedded_Apostrophe = TRUE;
    Ignore_Dollar = TRUE;
    Mode = Mode_Nroff;
    TeX_quotes = TRUE;
}


static void
#if NEW_STYLE
opt_output_file(
const char *filename
)
#else /* NOT NEW_STYLE */
opt_output_file(filename)
const char *filename;
#endif /* NEW_STYLE */
{
    Stdout = freopen(filename,"w",stdout);
    if (Stdout == (FILE*)NULL)
    {
	(void)fprintf(Stderr,
		      "\n%s open failure on output file [%s]\n",
		      ERROR_PREFIX, filename);
	exit(EXIT_FAILURE);
    }
}


static void
opt_page(VOID)
{
    Ignore_Formfeed = FALSE;
}


static void
opt_paragraph(VOID)
{
    Paragraph_Reset = TRUE;
}


static void
opt_Scribe(VOID)
{
    initialize_startup_defaults();
    Ignore_At = FALSE;
    Ignore_Backslash = TRUE;
    Ignore_Embedded_Apostrophe = TRUE;
    Mode = Mode_Scribe;
    TeX_quotes = TRUE;
}


static void
opt_TeX(VOID)
{
    initialize_startup_defaults();
    Ignore_Backslash = FALSE;
    Ignore_Dollar = FALSE;
    Ignore_Embedded_Apostrophe = TRUE;
    Mode = Mode_TeX;
    TeX_quotes = TRUE;
}


static void
opt_TeXinfo(VOID)
{
    opt_Scribe();
    Mode = Mode_TeXinfo;
}


static void
opt_version(VOID)
{
    show_version();
    exit(EXIT_SUCCESS);
}


static void
#if NEW_STYLE
out_strings(FILE *fpout, const char *lines[])
#else
out_strings(fpout, lines)
FILE *fpout;
const char *lines[];
#endif
{
    size_t k;

    for (k = 0; lines[k] != (const char*)NULL; k++)
	(void)fputs(lines[k], fpout);
}


static void
#if NEW_STYLE
pop(Stack_t *stack, int expect)
#else /* NOT NEW_STYLE */
pop(stack,expect)
Stack_t *stack;
int expect;
#endif /* NEW_STYLE */
{
    if (!Nested && (stack == &Dstack))
	return;
    if (stack->top == 0)
    {
	(void)fprintf(Stdout,"%s:delimiter stack underflow: expected `%c' on stack\n",
		     get_position((Stack_t*)NULL,0,Verbose), expect);
	Error_Count++;
    }
    else if (stack->stack[--stack->top].delimiter != expect)
    {
	char posbuf[POSITION_SIZE];

	(void)strncpy (&posbuf[0],
		       get_position((Stack_t*)NULL,0,Verbose), sizeof (posbuf));
	(void)fprintf(Stdout,"%s:delimiter mismatch: saw `%c' at %s: expected `%c'\n",
		     posbuf,
		     stack->stack[stack->top].delimiter,
		     get_position(stack,stack->top,Compact),
		     expect);
	Error_Count++;
    }
}


static void
#if NEW_STYLE
push(Stack_t *stack, int c)
#else /* NOT NEW_STYLE */
push(stack,c)
Stack_t *stack;
int c;
#endif /* NEW_STYLE */
{
    if (!Nested && (stack == &Dstack))
	return;
    if (stack->top > (size_t)(MAX_STACK - 1))
    {
	(void)fprintf(Stdout,"%s:delimiter stack overflows limit of %d\n",
		     get_position((Stack_t*)NULL,0,Verbose),
		     (int)MAX_STACK);
	Error_Count++;
    }
    else
    {
	Current_Position.delimiter = c;
	(void)memcpy((VOIDP)&stack->stack[stack->top],
		     (VOIDP)&Current_Position,
		     sizeof(Position_t));
	stack->top++;
    }
}


static void
#if NEW_STYLE
reset(size_t keep)
#else /* NOT NEW_STYLE */
reset(keep)
size_t keep;
#endif /* NEW_STYLE */
{
    while ((Dstack.top > 0) && (Dstack.top > keep))
    {
	--Dstack.top;
	(void)fprintf(Stdout,"%s:unmatched delimiter `%c'\n",
		     get_position(&Dstack,Dstack.top,Verbose),
		     Dstack.stack[Dstack.top].delimiter);
	Error_Count++;
    }
    initialize_flags();
}


static int
#if NEW_STYLE
set_position(int c)
#else /* NOT NEW_STYLE */
set_position(c)
int c;
#endif /* NEW_STYLE */
{
    if (c != EOF)
    {
	if (c == CHAR_NEWLINE)
	    Last_Newline_Position = Current_Position;

	Current_Position.char_number++;
	Current_Position.delimiter = c;

	if (c == CHAR_NEWLINE)
	{
	    Current_Position.column_number = 0;
	    Current_Position.line_number++;
	}
	else if (c == CHAR_HT)
	    Current_Position.column_number =
	      (Current_Position.column_number + 7) & ~7;
	else
	    Current_Position.column_number++;
    }
    return (c);
}


static void
show_version(VOID)
{
    static const char *version_text[] =
    {
	"This is ", PROGRAM, " version ", VERSION, "\n",
	"\n",
	(const char *)NULL,
    };

    out_strings(Stderr, version_text);
}


static int
#if NEW_STYLE
strnicmp(
const char	*s1,
const char	*s2,
size_t		n
)
#else
strnicmp(s1,s2,n)
const char	*s1;
const char	*s2;
size_t		n;
#endif
{
    int	   c1;
    int	   c2;

    /*******************************************************************
    Compare strings ignoring case, stopping after n characters, or at
    end-of-string, whichever comes first.
    *******************************************************************/

    for (; (n > 0) && *s1 && *s2; ++s1, ++s2, --n)
    {
	c1 = 0xff & (islower((int)*s1) ? (int)*s1 : tolower((int)*s1));
	c2 = 0xff & (islower((int)*s2) ? (int)*s2 : tolower((int)*s2));
	if (c1 < c2)
	    return (-1);
	else if (c1 > c2)
	    return (1);
    }
    if (n == 0)		   /* first (original) n characters match */
	return (0);
    else if (*s1 == (char)CHAR_NUL)
	return ((*s2 == (char)CHAR_NUL) ? 0 : -1);
    else /* (*s2 == (char)CHAR_NUL) */
	return (1);
}


static int
#if NEW_STYLE
unput(int c)
#else /* NOT NEW_STYLE */
unput(c)
int c;
#endif /* NEW_STYLE */
{
    if ((c != EOF) && (Input_Pushback_Count < MAX_PUSHBACK))
    {
	Input_Pushback_Buffer[Input_Pushback_Count++] = c;
	if (c == CHAR_NEWLINE)	/* this works as long as lookahead is never */
	    Current_Position = Last_Newline_Position; /* more than one newline */
	else
	{
	    Current_Position.char_number--;
	    Current_Position.delimiter = c;
	    Current_Position.column_number--; /* may be incorrect at tab */
	}
    }
    return (c);
}


static void
#if NEW_STYLE
usage(const char *bad_option)
#else /* NOT NEW_STYLE */
usage(bad_option)
const char *bad_option;
#endif /* NEW_STYLE */
{
    static const char *usage_text[] =
    {
	"Usage:\n",
	"\t",
	(const char*)NULL,	/* replaced by Program_Name at run time */
	"\t[-?] [-author] [-BibTeX] [-copyright] [-help]\n",
	"\t\t\t[-error-file filename] [-ignore delimiters] [-Lisp]\n",
	"\t\t\t[-nested] [-no-BibTeX] [-no-Lisp] [-no-nested]\n",
	"\t\t\t[-no-Nroff] [-no-page-reset] [-no-paragraph-reset]\n",
	"\t\t\t[-no-Scribe] [-no-TeX] [-no-TeXinfo] [-Nroff]\n",
	"\t\t\t[-output-file filename] [-page-reset] [-paragraph-reset]\n",
	"\t\t\t[-Scribe] [-TeX] [-TeXinfo] [-version]\n",
	"\t\t\t[files or <infile] [>outfile]\n",
	(const char*)NULL,
    };

    show_version();
    if (bad_option != (const char*)NULL)
	(void)fprintf(Stderr, "Unrecognized option [%s]\n", bad_option);
    usage_text[2] = Program_Name;
    out_strings(Stderr, usage_text);
}
